/*
 *  RecAgent.h
 *  
 *  Created by Jordan Frank on Thu December 7 2006.
 *
 *  Description: 
 *  
 *	Designed for the PudleWorld.cpp environment
 *
 *  This agent uses off-policy temporal-difference learning with function approximation
 *  and recognizers, from [1].
 * 
 *  Uses CMAC tile coding based on TileAgent.cpp from RL-Glue 
 *
 *	State: Continous valued - 3D
 *	Actions: Continous Valued - 2D
 *
 *  [1] - Precup, D., Sutton, R. S., Paduraru, C., Koop, A., Singh, S., 
 *        Off-policy Learning with Recognizers, 2006.
 */

#include "RecAgent.h" 
#include "data.c"

void agent_init(Task_specification task_spec)
{
    int i;
    task_spec_struct ps;
    parse_task_spec(task_spec, &ps);
    stateDim = (int)ps.obs_maxs[0]+1;
    numActions = (int)ps.action_maxs[0]+1;
    
    fprintf(stderr,"\ntask_spec: %s, state_dims: %d, num_actions: %d\n",
            task_spec,stateDim,numActions);

    //Initalize all data
    curr_state = new double[stateDim];
    prev_state = new double[stateDim];
    theta = new double[memorysize];
    non_rec_theta = new double[memorysize];
    
    mu_weights = new double[memorysize];

    e = new double[memorysize];
    non_rec_e = new double[memorysize];

    featurevec = new int[tilings];
    input = new float[stateDim];

    step = 0;
    srand(999);
    openRecProbFiles();
    openRewEstFiles();

    recognized = 0;
    total = 0;
}

Action agent_start(Observation o)
{
    int i;
    t = 0;
    curr_action = randomAction();
    for(int i = 0; i < stateDim; i++)
        curr_state[i] = o[i];
    getFeatureVector(curr_state, featurevec);

    k = g;
    non_rec_k = g;
    for (i = 0; i < tilings; i++)
    {
        e[featurevec[i]] = k;
        non_rec_e[featurevec[i]] = non_rec_k;        
    }
    printRecProbs();
    printRewEst();
    
    return curr_action;
}

Action agent_step(Reward rew, Observation o)
{
    double delta, sum, mu, rho;
    int prev_recognized, curr_recognized;
    int curr_featurevec[tilings];
    double yp, y, dy, gamma;
    int i;

    t++;
    // Decay the learning rates.
    decayLearningRates(t);

    // Save the previous action and state
    prev_action = curr_action;
    for(i = 0; i < stateDim; i++)
        prev_state[i] = curr_state[i];

    // Store the current state and generate the next action based
    // on our behaviour policy (which is random for now)
    for(i = 0;i < stateDim; i++)
        curr_state[i] = o[i];

    curr_action = randomAction(); // Behaviour policy is random.
//    curr_action = favourUpRightAction();
    
//    prev_recognized = recognize(&recognize_up_right, prev_state, prev_action);
//    curr_recognized = recognize(&recognize_up_right, curr_state, curr_action);
    prev_recognized = recognize(&recognize_good_actions, prev_state, prev_action);
    curr_recognized = recognize(&recognize_good_actions, curr_state, curr_action);

    mu = getMu(featurevec);
    updateMu(featurevec, prev_recognized);
    y = vecvecMult(featurevec, theta, tilings);

    getFeatureVector(curr_state, curr_featurevec);
    yp = vecvecMult(curr_featurevec, theta, tilings);

    // Calculate using estimated recognition probability.
    rho = prev_recognized / mu;
    delta = rho * (rew + (1.0 - beta) * yp - y);
    for (i = 0; i < memorysize; i++)
    {
        theta[i] += alpha * delta * e[i];
    }
    k = curr_recognized * k + g;
    for (i = 0; i < memorysize; i++)
    {
        e[i] = (1.0 - beta) * lambda * e[i];
    }
    for (i = 0; i < tilings; i++)
    {
        e[curr_featurevec[i]] += k;
    }
    for (i = 0; i < memorysize; i++)
    {
        if (e[i] > 1.0) e[i] = 1.0;
    }

    y = vecvecMult(featurevec, non_rec_theta, tilings);
    yp = vecvecMult(curr_featurevec, non_rec_theta, tilings);

    // Calculate using exact recognition probability.
    // MAKE SURE THIS IS UPDATED IF RECOGNIZER IS CHANGED!!!
    mu = getExactComplexRecProb(prev_state);
    rho = prev_recognized / mu;
    delta = rho * (rew + (1.0 - beta) * yp - y);
    for (i = 0; i < memorysize; i++)
    {
        non_rec_theta[i] += alpha * delta * non_rec_e[i];
    }
    non_rec_k = curr_recognized * non_rec_k + g;
    for (i = 0; i < memorysize; i++)
    {
        non_rec_e[i] = (1.0 - beta) * lambda * non_rec_e[i];
    }
    for (i = 0; i < tilings; i++)
    {
        non_rec_e[curr_featurevec[i]] += non_rec_k;
    }
    for (i = 0; i < memorysize; i++)
    {
        if (non_rec_e[i] > 1.0) non_rec_e[i] = 1.0;
    }
    
    for (i = 0; i < tilings; i++)
    {
        featurevec[i] = curr_featurevec[i];
    }

    printRecProbs();
    printRewEst();
    
    return curr_action;
}

void agent_end(Reward r)
{
    // Continuous task, this is never called.
}

void agent_cleanup()
{
    double p;
    
    //clean up memory
    printMus();
    printRew();
    closeRecProbFiles();
    closeRewEstFiles();
    p = (float)recognized / (float)total;
    
    printf("Recognized: %d out of %d total (%f).\n",recognized,total,p);
//    printf("Yet our estimate for mu is .\n",
    delete [] curr_state;
    delete [] prev_state;
    delete [] featurevec;
    delete [] input;    

    delete [] theta;
    delete [] non_rec_theta;
    
    delete [] mu_weights;
    
    delete [] non_rec_e;
    delete [] e;
}

void decayLearningRates(int t)
{
    if (t==T || t==3*T || t==7*T || t==15*T ||
        t==31*T || t==63*T || t==127*T)
    {
        alpha = alpha / 2.0;
        mu_alpha = mu_alpha / 2.0;
    }
}

void getFeatureVector(Observation _o, int* phi)
{
    // Get the feature vector from our tile coding scheme.
    float input[stateDim];
    for(int x=0;x<stateDim;x++)
        input[x] = _o[x]/tilewidth;
    GetTiles(phi,tilings,memorysize,input,stateDim);
}

Action randomAction()
{
    return rand()%numActions;
}

Action favourUpRightAction()
{
    int r;
    Action a;
    r = rand()%100;
    a = rand()%4;
    // 40% chance of first four actions
    if (r < 40) return a;
    // 25% chance of second four actions
    else if (r < 65) return a + 4;
    // 10% chance of third four actions
    else if (r < 75) return a + 8;
    // 25% chance of last four actions
    else return a + 12;
}


int recognize(Recognizer rec, Observation _o, Action _a)
{
    return rec(_o, _a);
}

int recognize_up_right(Observation _o, Action _a)
{
    total++;
    // Returns 1 if an action is recognized or 0 otherwise.
    // Right now the actions that we recognize do not depend on the state.

    // Simple recognizer, recognizes actions that take us up or to the 
    // right.
    if (0 <= _a && _a < 4)
    { 
        // Recognize all actions heading up and to the right.
        recognized++;
        return 1;
    }
    else
    {
        return 0;
    }
}

int recognize_good_actions(Observation _o, Action _a)
{
    total++;
    // If we are in the lower left quadrant of the world, then
    // recognize the 5 actions that move us up and to the right.
    if (_o[0] < 0.5 && _o[1] < 0.5) 
    {
        if (_a <= 4)
        {
            recognized++;
            return 1;
        }
        else
        {
            return 0;
        }
    }    
    // If we are in the upper left quadrant of the world, then
    // recognize the 7 actions that move us to the right.
    else if (_o[0] < 0.5 && _o[1] >= 0.5) 
    {
        if (0 < _a && _a < 8)
        {
            recognized++;
            return 1;
        }
        else
        {
            return 0;
        }
    }

    // If we are in the lower right quadrant of the world, then
    // recognize the 7 actions that move us upwards.
    else if (_o[0] >= 0.5 && _o[1] < 0.5)
    {
        if ((_a < 4) || (12 < _a))
        {
            recognized++;
            return 1;
        }
        else
        {
            return 0;
        }
    }    

    // Lastly, if we are in the upper right quadrant of the world,
    // then the recognized actions are the 5 that move us towards
    // the center of the goal state.
    else 
    {
        if (_o[0] <= 0.85 && _o[1] <= 0.85)
        {
            if (_a <= 4)
            {
                recognized++;
                return 1;
            }
            else return 0;
        }
        else if (_o[0] <= 0.85 && _o[1] > 0.85)
        {
            if (4 <= _a && _a <= 8)
            {
                recognized++;
                return 1;
            }
            else return 0;
        }
        else if (_o[0] > 0.85 && _o[1] > 0.85)
        {
            if (8 <= _a && _a <= 12)
            {
                recognized++;
                return 1;
            }
            else return 0;
        }
        else
        {
            if (12 <= _a)
            {
                recognized++;
                return 1;
            }
            else return 0;
        }        
    }
}

double getExactComplexRecProb(Observation _o)
{
    // If we are in the lower left quadrant of the world, then
    // recognize the 5 actions that move us up and to the right.
    if (_o[0] < 0.5 && _o[1] < 0.5) 
    {
        return 0.3125;
    }    
    // If we are in the upper left quadrant of the world, then
    // recognize the 7 actions that move us to the right.
    else if (_o[0] < 0.5 && _o[1] >= 0.5) 
    {
        return 0.4375;
    }

    // If we are in the lower right quadrant of the world, then
    // recognize the 7 actions that move us upwards.
    else if (_o[0] >= 0.5 && _o[1] < 0.5)
    {
        return 0.4375;
    }

    // Lastly, if we are in the upper right quadrant of the world,
    // then the recognized actions are the 5 that move us towards
    // the center of the goal state.
    else 
    {
        return 0.3125;
    }
}

double getMu(int *features)
{
    double mu;
    double wx;
    wx = vecvecMult(features, mu_weights, tilings);
    mu = sigmoid(wx);
    return mu;
}

void updateMu(int *features, double recognized)
{
    double delta;
    int i;
    delta = recognized - getMu(features);
    delta = mu_alpha * delta;
    
    // printf("mu was: %f.\n", getMu(features));
    for (i = 0; i < tilings; i++)
    {
        mu_weights[features[i]] = mu_weights[features[i]] + delta;
    }
    // printf("mu is now: %f. ", getMu(features));
}

double vecvecMult(int *a, double *b, int size) 
{
    int i;
    double sum = 0.0;
    for (i = 0; i < size; i++)
    {
        sum += b[a[i]];
    }
    return sum;
}

double sigmoid(double v)
{
    return 1.0 / (1 + exp(-v));
}
